/**
 *
 * Copyright 2005-2006 The Apache Software Foundation
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.logicblaze.soa;

import java.io.File;
import java.io.InputStream;
import java.io.PrintStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.JarURLConnection;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * Main class bootstraps the classpath. This could be part of XBean
 * 
 * @version $Revision: 4528 $
 */
public class Main {

    public static final String TASK_DEFAULT_CLASS = "org.logicblaze.soa.BootstrapMain";

    private File home;
    private ClassLoader parentClassLoader = Main.class.getClassLoader();
    private ClassLoader classLoader;
    private List extensions = new ArrayList(10);
    private boolean verbose;
    public String mainClassName = TASK_DEFAULT_CLASS;
    private Object mainObject;


    private static boolean useDefExt = true;

    public static void main(String[] args) {
        Main app = new Main();

        // Convert arguments to collection for easier management
        List tokens = new LinkedList(Arrays.asList(args));
        // Parse for extension directory option
        app.parseExtensions(tokens);

        // Add default extension directories
        /*
         * if(useDefExt && app.canUseExtdir()) {
         */
        if (useDefExt && app.canUseExtdir()) {
            File shared = new File(app.getHome(), "shared");
            app.addExtensionDirectory(new File(shared, "lib"));
            app.addExtensionDirectory(new File(shared, "classes"));
        } else {
            if( app.isVerbose() ) {
                System.out.println("Extension directories were not added");
            }
        }

        try {
            app.main(tokens);
        }
        catch (Throwable e) {
            System.out.println("Failed to execute main task. Reason: " + e);
            e.printStackTrace();
        }
    }

    public void parseExtensions(List tokens) {
        if (tokens.isEmpty()) {
            return;
        }

        int count = tokens.size();
        int i = 0;

        // Parse for all --extdir and --noDefExt options
        while (i < count) {
            String token = (String) tokens.get(i);
            // If token is an extension dir option
            if (token.equals("--extdir")) {
                // Process token
                count--;
                tokens.remove(i);

                // If no extension directory is specified, or next token is
                // another option
                if (i >= count || ((String) tokens.get(i)).startsWith("-")) {
                    System.out.println("Extension directory not specified.");
                    System.out.println("Ignoring extension directory option.");
                    continue;
                }

                // Process extension dir token
                count--;
                File extDir = new File((String) tokens.remove(i));

                /*
                 * if(!canUseExtdir()) { System.out.println("Extension directory
                 * feature not available due to the system classpath being able
                 * to load: " + taskClass);
                 * System.out.println("Ignoring extension directory option.");
                 * continue; }
                 */

                if (!extDir.isDirectory()) {
                    System.out.println("Extension directory specified is not valid directory: " + extDir);
                    System.out.println("Ignoring extension directory option.");
                    continue;
                }

                addExtensionDirectory(extDir);
            }
            else if (token.equals("--noDefExt")) { // If token is --noDefExt
                // option
                count--;
                tokens.remove(i);
                useDefExt = false;
            }
            else if (token.equals("-v") || token.equals("--verbose")) {
                count--;
                tokens.remove(i);
                verbose = true;
            }
            else {
                i++;
            }
        }

    }

    public void main(List tokens) throws Exception {
        System.out.println("====================================================");
        System.out.println("FUSE by LogicBlaze, Inc. Copyright (c) 2006");
        System.out.println("====================================================");
        System.out.println("For more information see: http://www.logicblaze.com/");
        System.out.println("For support please contact: http://support.logicblaze.com/");
        System.out.println();
        System.out.println("Fuse is starting up from install directory: " + getHome());

        ClassLoader cl = getClassLoader();
        String baseDirectory = getHome().getAbsolutePath();

        // Use reflection to run the task.
        try {
            
            if( mainObject == null ) {
                mainObject = cl.loadClass(mainClassName).newInstance();
            }
            
            String[] args = (String[]) tokens.toArray(new String[tokens.size()]);
            Method runTask = mainObject.getClass().getMethod("main", new Class[] { String[].class, InputStream.class, PrintStream.class, String.class });
            runTask.invoke(mainObject, new Object[] { args, System.in, System.out, baseDirectory });
        }
        catch (InvocationTargetException e) {            
            Throwable cause = e.getCause();
            if( cause instanceof Exception )
                throw (Exception)cause;
            throw new RuntimeException(cause.getMessage(), cause);
            
        }
    }
    
    public void addExtensionDirectory(File directory) {
        extensions.add(directory);
    }

    /**
     * The extension directory feature will not work if the broker factory is
     * already in the classpath since we have to load him from a child
     * ClassLoader we build for it to work correctly.
     * 
     * @return
     */
    public boolean canUseExtdir() {
        try {
            parentClassLoader.loadClass(mainClassName);
            if (verbose) {
                System.out.println("Cannot use extension mechanism since "+mainClassName+" could be loaded from the parent ClassLoader.");
            }
            return false;
        }
        catch (ClassNotFoundException e) {
            return true;
        }
    }

    public ClassLoader getClassLoader() throws MalformedURLException {
        if (classLoader == null) {
            // Setup the ClassLoader
            classLoader = parentClassLoader;
            if (!extensions.isEmpty()) {

                ArrayList urls = new ArrayList();
                for (Iterator iter = extensions.iterator(); iter.hasNext();) {
                    File dir = (File) iter.next();
                    if (verbose) {
                        try {
                            System.out.println("Adding to classpath: " + dir.getCanonicalPath());
                        }
                        catch (Exception e) {
                        }
                    }
                    urls.add(dir.toURL());
                    File[] files = dir.listFiles();
                    if (files != null) {
                        for (int j = 0; j < files.length; j++) {
                            if (files[j].getName().endsWith(".zip") || files[j].getName().endsWith(".jar")) {
                                if (verbose) {
                                    try {
                                        System.out.println("Adding to classpath: " + files[j].getCanonicalPath());
                                    }
                                    catch (Exception e) {
                                    }
                                }
                                urls.add(files[j].toURL());
                            }
                        }
                    }
                }

                URL u[] = new URL[urls.size()];
                urls.toArray(u);
                classLoader = new URLClassLoader(u, classLoader) {
                    public String toString() {
                        return super.toString() + "[uris=" + Arrays.asList(getURLs()) + "]";
                    }
                };
                
            } else {
                if (verbose) {
                    System.out.println("No extension directories to add.");
                }
            }            
            Thread.currentThread().setContextClassLoader(classLoader);
        }
        return classLoader;
    }

    public void setHome(File activeMQHome) {
        this.home = activeMQHome;
    }

    public File getHome() {
        if (home == null) {
            if (System.getProperty("fuse.home") != null) {
                home = new File(System.getProperty("fuse.home"));
            }

            if (home == null) {
                // guess from the location of the jar
                URL url = Main.class.getClassLoader().getResource("org/logicblaze/soa/Main.class");
                if (url != null) {
                    try {
                        JarURLConnection jarConnection = (JarURLConnection) url.openConnection();
                        url = jarConnection.getJarFileURL();
                        URI baseURI = new URI(url.toString()).resolve(".");
                        home = new File(baseURI).getCanonicalFile();
                    }
                    catch (Exception ignored) {
                    }
                }
            }

            if (home == null) {
                home = new File(".");
            }
        }
        return home;
    }

    public boolean isVerbose() {
        return verbose;
    }

    public void setVerbose(boolean verbose) {
        this.verbose = verbose;
    }

    public ClassLoader getParentClassLoader() {
        return parentClassLoader;
    }

    public void setParentClassLoader(ClassLoader parentClassLoader) {
        this.parentClassLoader = parentClassLoader;
    }

    public String getMainClassName() {
        return mainClassName;
    }

    public void setMainClassName(String taskClass) {
        this.mainClassName = taskClass;
    }

}
